/*
Copyright (C) 2003 by Kristis Makris. See the file AUTHORS for contact
information of all contributors.

This file is part of a framework that allows dynamic replacement of
core kernel components.

You may not use this work in any format (source/binary/documentation,
electronic/hardcopy.) for any purpose whatsoever, without the
permission of Kristis Makris. At some point in the future, this
program will be free software, licensed under the GPL License.
*/



/* This is a benchmark library used to time how long it takes for a
   user defined operation to complete. It's a collection of routines
   that facilitate the activity of gathering such a measurements. */
#include "benchmark_lib_core.h"
#include "devel.h"


/** Flags whether the benchmark test failed for any reason. */
short benchmark_test_tainted = 0;

/** Total number of observations that can be stored */
unsigned long benchmark_observations_number = -1;

/** Number of observations stored so far. */
unsigned long benchmark_observations_counter = 0;

/** Flags if the observation buffer is full. */
short benchmark_observations_buffer_full = 0;

/** State of benchmark library. Can be enabled or disabled. */
short benchmark_state = 0;

/** Guards access to the benchmark observations buffer and the benchmark_state variable */
benchmark_lock_t benchmark_lock;

/** Used to store the time taken to run a certain activity, as defined
    by the user implementation. Pointer to an array. */
benchmark_times_t (*benchmark_observations)[];



/** \brief Used to signal that there was an error during the
    benchmarking activity.
*/
void benchmark_flag_error(void)
{
  benchmark_test_tainted = 1;
}


/** \brief Returns the state of the benchmark library. 
    \note Requires holding the benchmark_lock.
    \retval  1 if enabled,
    \retval  o otherwise.
*/
int benchmark_lib_enabled(void)
{
  return (benchmark_state == 1);
}



/** \brief Enables the benchmark library.

   \note Requires holding the benchmark_lock. */
void benchmark_lib_enable(void)
{
  benchmark_state = 1;
}



/** \brief Enables the benchmark library. 

    \note Requires holding the benchmark_lock. */
void benchmark_lib_disable(void)
{
  benchmark_state = 0;
}



/** \brief Initializes the lock to the benchmark observations
    buffer */
void benchmark_lib_lock_init(void)
{
  spin_lock_init(&benchmark_lock);
}



/** \brief Grabs the lock to the observations buffer. */
void inline benchmark_lib_lock(void)
{
  spin_lock(&benchmark_lock);
}



/** \brief Releases the lock to the observations buffer. */
void inline benchmark_lib_unlock(void)
{
  spin_unlock(&benchmark_lock);
}


/** \brief Initializes the benchmarking library.
    \param observations_number The maximum number of observations that
           will be collected.
 */
void benchmark_lib_init(unsigned long observations_number)
{
  /* Initialize the lock to the benchmark observations buffer. This
     should happen only once. */
  benchmark_lib_lock_init();

  /* Initialize the internal state of the benchmark library */
  benchmark_lib_init_internal(observations_number);
}



/** \brief Initializes the internal state of the benchmark library */
void benchmark_lib_init_internal(unsigned long observations_number)
{

  benchmark_test_tainted = 0;
  benchmark_observations_counter = 0;
  benchmark_observations_number = observations_number;

  /* Allocate enough memory for the entire array of elements */
  benchmark_observations = (benchmark_times_t (*)[]) devel_malloc(sizeof(benchmark_times_t) * benchmark_observations_number);

  if (benchmark_observations != NULL) {
    /* Zero out all array values */
    memset(benchmark_observations, 0, sizeof(benchmark_times_t) * benchmark_observations_number);

    /* Start with a clean set of benchmark values. */
    memset(benchmark_observations, 0, sizeof(sizeof(benchmark_times_t) * benchmark_observations_number));
  } else {
    devel_debug_print( DEVEL_DEBUG_NORMAL, BENCHMARK_LIB_NAME ": benchmark_lib_init_internal ran out of memory and could not reinitialize. 0x%lx bytes were requested.\n", sizeof(benchmark_times_t) * benchmark_observations_number);
  }
}



int benchmark_lib_module_init(void)
{
  int retval = 0;

  devel_debug_print( DEVEL_DEBUG_NORMAL, BENCHMARK_LIB_NAME ": version " BENCHMARK_LIB_VERSION_STR "\n");

  /* Register the file operations */
  //  retval = benchmark_lib_file_operations_init();
  retval = 0;

  devel_debug_print(DEVEL_DEBUG_NORMAL, BENCHMARK_LIB_NAME ": finished initialization\n");

  return retval;
}



/*! Uninitializes the benchmark library */
void benchmark_lib_exit(void)
{
}



/* Uninitializes the internal state of the benchmark library */
void benchmark_lib_exit_internal(void)
{
  if (benchmark_observations != NULL)
    devel_free(benchmark_observations);
}



/* Unitializes the state of the benchmark library as a module */
void benchmark_lib_module_exit(void)
{
  /* Remove the file operations interface */
  //benchmark_lib_file_operations_cleanup();

  benchmark_lib_exit_internal();
  devel_debug_print(DEVEL_DEBUG_NORMAL, BENCHMARK_LIB_NAME ": finished cleanup\n");
}



/* Resets the benchmark library */
void benchmark_lib_reset(void)
{
  benchmark_lib_exit_internal();
  benchmark_lib_init_internal(benchmark_observations_number);
  benchmark_observations_buffer_full = 0;
}



/*! Stores the difference between the two benchmark times in the
   observations buffer.
   \param time_before The time before the user defined activity started
   \param time_after The time after the user defined activity ended
*/
void benchmark_lib_store_times(benchmark_times_t *time_before, benchmark_times_t *time_after)
{
  benchmark_times_t time_diff;
  devel_debug_print(DEVEL_DEBUG_TRACE, "benchmark_store_times: entering\n");

  /* Find the difference between the two time intervals */
  benchmark_lib_times_diff(time_after, time_before, &time_diff);

  /* Lock access to the observations buffer */
  benchmark_lib_lock();

  if (benchmark_lib_enabled()) {
    /* Make sure the buffer is not overrun */
    if (benchmark_observations_counter < benchmark_observations_number) {
      /* Copy the time observation */
      devel_debug_print(DEVEL_DEBUG_DETAIL, "benchmark_store_times: Storing times %ld and %ld\n", time_diff.benchmark_times_seconds, time_diff.benchmark_times_microseconds);

      memcpy(&(*benchmark_observations)[benchmark_observations_counter], &time_diff, sizeof(benchmark_times_t));
      benchmark_observations_counter++;
    } else {
      /* Emit a log message if the buffer is full */
      if (benchmark_observations_buffer_full == 0 &&
	  benchmark_observations_counter >= benchmark_observations_number) {
	devel_debug_print(DEVEL_DEBUG_STATE, "benchmark_store: observation buffer is full with %ld entries\n", benchmark_observations_counter);
	benchmark_observations_buffer_full = 1;
      }     
    }
  }

  /* Release the lock to the observations buffer */
  benchmark_lib_unlock();

  devel_debug_print(DEVEL_DEBUG_TRACE, "benchmark_store_times: exiting\n");
}



/* Returns a string representation of the data collected during
   benchmarking. The string pointer returned is dynamically allocated,
   and has to be freed by the caller. */
char * benchmark_lib_tostring(void)
{
  long data_size = 1024 * 8;
  char *data;
  char data_tmp[data_size / 8];
  unsigned long counter = 0;

  data = (char *) devel_malloc(data_size);
  memset(data, 0, data_size);
  memset(data_tmp, 0, data_size / 8);

  /* Grab the lock to the observations buffer */
  benchmark_lib_lock();

  /* Display the version of the library */
  strcpy(data, BENCHMARK_LIB_NAME " " BENCHMARK_LIB_VERSION_STR "\n");
  /* STRCPY, since its the first copy */

  /* Display some critical information such as if the benchmarks were
     tainted and the number of total observations collected. */
  sprintf(data_tmp, "Tainted: %d\n", benchmark_test_tainted);
  strcat(data, data_tmp);
  sprintf(data_tmp, "Number of current observations: %ld\n", benchmark_observations_counter);
  strcat(data, data_tmp);
  sprintf(data_tmp, "Number of maximum observations: %ld\n", benchmark_observations_number);
  strcat(data, data_tmp);

  /* Display a header of what we are outputing and then show all
     observation results collected. */
  strcat(data, "Number\tSeconds\tMicroseconds\n");

  /* Make sure we have been initialized */
  if (benchmark_observations == NULL)
    goto out;

  while (counter < benchmark_observations_number ) {
    /* Observation number */
    sprintf(data_tmp, "%ld", counter + 1); /* +1 since the first
					      observation has index 0
					      in the observations
					      array */
    strcat(data, data_tmp);

    /* Seconds */
    sprintf(data_tmp, " %ld", (unsigned long) (*benchmark_observations)[counter].benchmark_times_seconds);
    strcat(data, data_tmp);

    /* Microseconds */
    sprintf(data_tmp, " %ld", (unsigned long) (*benchmark_observations)[counter].benchmark_times_microseconds);
    strcat(data, data_tmp);

    /* Next line*/
    strcat(data, "\n");

    counter++;
  }

 out:
  /* Release the lock to the observations buffer */
  benchmark_lib_unlock();

  return data;
}



/*! Computes the difference between big and small. It computes big - small.
 * \param big Pointer to a struct timeval timer. Should be the old timer,
 *            the bigger timer value.
 * \param small Pointer to a struct timeval timer. Should be the new timer,
 *              the smaller timer value.
 * \param diff Pointer to a struct timeval in which the result of the 
 *             operation big - small will be stored.
 * \bug The function does not check to ensure that big is indeed bigger than
 *      small
 * \note 1 second = 1000 milliseconds = 1000000 microseconds
 */
void benchmark_lib_times_diff(benchmark_times_t *big,
			      benchmark_times_t *small,
			      benchmark_times_t *diff)
{
  /* Subtracting the seconds is easy */
  diff->benchmark_times_seconds = big->benchmark_times_seconds - small->benchmark_times_seconds;

  /* If the benchmark_times_microseconds in big are greater than small, we are ok */
  if (big->benchmark_times_microseconds >= small->benchmark_times_microseconds)
    diff->benchmark_times_microseconds = big->benchmark_times_microseconds - small->benchmark_times_microseconds;
  else
    /* Carry over 1000000 microseconds in our subtraction */
    {
      diff->benchmark_times_microseconds = 1000000 + big->benchmark_times_microseconds - small->benchmark_times_microseconds;
      diff->benchmark_times_seconds -= 1; /* 1 second = 1000000 microseconds */
    }
}


#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include <sys/types.h>
#include <unistd.h>

static struct tms tms_beg;
static struct tms tms_end;
static int has_times = 0;


void benchmark_lib_pstats_init()
{
    has_times = times(&tms_beg) != -1;
}

void benchmark_lib_ps_report()
{
    char buf[1024];
#ifdef __APPLE__ /*  MAC OS X */
    sprintf(buf,"ps -o command,majflt,minflt,rss,pagein,vsz -p %d 1>&2",getpid() );
#else /* GNU/Linux */
    sprintf(buf,"ps -o comm,majflt,minflt,rss,drs,pagein,sz,trs,vsz %d 1>&2",getpid() );
#endif    
    system( buf );
}


void benchmark_lib_tms_report()
{
    double cputime;
    if (! has_times )
        return;
    times(&tms_end);
    cputime = ( ((float)tms_end.tms_utime + tms_end.tms_stime + tms_end.tms_cutime + tms_end.tms_cstime ) -
                ((float)tms_beg.tms_utime + tms_beg.tms_stime + tms_beg.tms_cutime + tms_beg.tms_cstime ) )
               / sysconf(_SC_CLK_TCK);
    fprintf(stderr,"\tcputime=%.3f\n" , cputime);
}
