#!/usr/bin/perl

use strict;
use Data::Dumper;
use Time::HiRes qw(usleep ualarm gettimeofday tv_interval);
use File::Copy;
use File::Remove qw(remove);

my $put_in_memory = 1;

my $coming_from_memory = 0;
my $storage_type = "";
if ( $coming_from_memory ) {
    $storage_type = "memory";
} else {
    $storage_type = "harddisk";
}

my $memory_filesystem_path = "/dev/shm";
my $hostname;

main ( @ARGV );



sub main {
    my ( @arguments ) = ( @_ );

    if (scalar @arguments != 3) {
	print "usage: $0 <server_hostname> <server_version> <server_compiler>\n";
	exit -1;
    }

    my ( $server_ip, $server_version, $server_compiler ) = ( @arguments );
    $hostname=`uname -n | sed -e 's/\..*//'`;

    `rm -rf build/downloads`;
    if ( $put_in_memory ) {
        # Will download in a memory filesystem
        mkdir $memory_filesystem_path . "/downloads";
        mkdir "build";
        `cd build; ln -s $memory_filesystem_path/downloads downloads`;
    } else {
        mkdir "build/downloads";
    }


    test_client_connection_time ( $server_ip, $server_version, $server_compiler );
    test_client_transfer_rate ( $server_ip, $server_version, $server_compiler );
}



sub find_median {
    my ( $times ) = ( @_ );
    my $counter = 0;
    my $size;
    my $median;
    my $median_filename;

    print Dumper( $times );

    # First count them
    foreach my $item ( sort keys %{ $times } ) {
        $counter++;
    }
    $size = $counter;

    # Put them in an array
    my @sorted_times;

    foreach my $item ( sort keys %{ $times } ) {
        push @sorted_times, $times->{ $item };
    }
    @sorted_times = sort( @sorted_times );

    my $median_counter;
    if ( $size % 2 == 1) {
        # Odd number
        $median_counter = ($size + 1) / 2;
    } else {
        # Even number
        $median_counter = $size / 2;
    }
    $median = @sorted_times[ $median_counter - 1 ];


    # Now find the filename of the median
    foreach my $item ( sort keys %{ $times } ) {
        if ( $times->{ $item } == $median ) {
            $median_filename = $item;
        }
    }

    print "size = $size\n";
    print "median_counter = $median_counter\n";
    print "median = $median\n";
    print "median_filename = $median_filename\n";


    return ($median, $median_filename);
}



sub test_client_connection_time {
    my ( $server_ip, $server_version, $server_compiler ) = ( @_ );
    my $cmd = "";

    my $curdir = `pwd`;
    $curdir =~ s/\n//;
    chdir "build/downloads";


    my $times;
    my $counter = 1;
    my $number_of_times = 11;

    while ( $counter <= $number_of_times ) {

        remove( "small_file*" );
        my $max_iterations = 1000;
        my $iteration = 1;

        # Measure the connection time
        #$cmd = 'ncftp -u anonymous -p anonymous -P 2021 ' . $server_ip . ' > /dev/null 2>&1 <<EOF
#quit
#EOF
#';

        # Wget seems to retrieve the directory listing sooner. I'm not
        # sure that's appropriate though. I'd rather we did not
        # involve the filesystem at all
#        $cmd = 'wget --user=anonymous --password=anonymous ftp://' . $server_ip . ':2021 > /dev/null 2>&1';
        $cmd = 'wget --user=anonymous --password=anonymous ftp://' . $server_ip . ':2021/small_file > /dev/null 2>&1';


        # Measure elapsed time
        my $t0 = [ gettimeofday() ];
        while ( $iteration <= $max_iterations ) {

            
            system( $cmd );
            $iteration++;
        }
        my $t1 = [ gettimeofday() ];
        my $t0_t1 = tv_interval $t0, $t1;
        my $connection_time = $t0_t1 / $max_iterations * 1000;
        my $result = "Time to transfer: $t0_t1 seconds\n" .
            "Number of transfers: $max_iterations\n" .
            "Transfer rate: " . $connection_time . " milliseconds/transfer\n";
        print $result;
        open ( RESULT_FILE, "> $curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.small_file_32.result.$counter" );
        print RESULT_FILE $result;
        close ( RESULT_FILE );

        # We will store the connection time
        $times->{ $counter } = $connection_time;

        $counter++;
    }

    # Find the median
    my ( $median_rate, $median ) = find_median( $times );

    copy ("$curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.small_file_32.result.$median", "$curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.small_file_32.result");

    chdir $curdir;
}



sub test_client_transfer_rate {
    my ( $server_ip, $server_version, $server_compiler ) = ( @_ );
    my $cmd = "";

    # 600MB file
    my $size = 300 * 1024 * 1024;

    my $curdir = `pwd`;
    $curdir =~ s/\n//;
    chdir "build/downloads";


    my $times;
    my $counter = 1;
    my $number_of_times = 11;
    while ( $counter <= $number_of_times ) {
        # Remove a previous file
        remove ( "large_file*" );

        # Measure the transfer rate of a large file
        $cmd = 'wget --user=anonymous --password=anonymous ftp://' . $server_ip . ':2021/large_file > ' . $curdir . '/results/' . $server_ip . '.' . $storage_type . '.' . $server_version . '.' . $server_compiler . '.large_file_' . $size . '.output.' . $counter . ' 2>&1 ';
        

        # Measure elapsed time
        my $t0 = [ gettimeofday() ];
        system( $cmd );
        my $t1 = [ gettimeofday() ];

        my $t0_t1 = tv_interval $t0, $t1;
        my $transfer_rate = ( $size / $t0_t1 / 1000000 );
        my $result = "Time to transfer: $t0_t1 seconds\n" .
            "Size of transfer: $size bytes\n" .
            "Transfer rate: " . $transfer_rate . " MB/sec \n";
        print $result;
        open ( RESULT_FILE, "> $curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.large_file_$size.result.$counter" );
        print RESULT_FILE $result;
        close ( RESULT_FILE );

        # We will store the transfer rate
        $times->{ $counter } = $transfer_rate;

        $counter++;
    }

    # Find the median
    my ( $median_rate, $median ) = find_median( $times );

    # Store the results
    copy ("$curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.large_file_$size.result.$median", "$curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.large_file_$size.result");
    copy ("$curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.large_file_$size.output.$median", "$curdir/results/$server_ip.$storage_type.$server_version.$server_compiler.large_file_$size.output");

    chdir $curdir;
}
